-- Trace Multiplexer for Nios II designs with multiple Nios II cores,
-- which have an off-chip trace port.
--
-- (C) Lauterbach Datentechnik GmbH 2007
--
-- $Id: tracemultiplexer.vhd 2309 2007-12-07 16:34:13Z irohloff $
--
library IEEE;
use IEEE.std_logic_1164.all;

entity TraceSldNode is
	generic (
		SLD_NODE_INFO : INTEGER := 16#08084C00#
	);
	port (
		clrn    : in STD_LOGIC := '0';
		raw_tck : in STD_LOGIC := '0';
		tdi     : in STD_LOGIC := '0';
		jtag_state_cdr : in STD_LOGIC := '0'; -- '1' in Capture-DR state
		jtag_state_sdr : in STD_LOGIC := '0'; -- '1' in Shift-DR state
		jtag_state_udr : in STD_LOGIC := '0'; -- '1' in Update-DR state
		tdo     : out STD_LOGIC;

		usr1    : in STD_LOGIC := '0';
		ena     : in STD_LOGIC := '0';
		ir_in   : in STD_LOGIC_VECTOR(1 downto 0) := (others => '0');
		ir_out  : out STD_LOGIC_VECTOR(1 downto 0);

		iClkX2         : in STD_LOGIC;

		iTraceClkCpu0  : in STD_LOGIC := '0';
		iTraceDataCpu0 : in STD_LOGIC_VECTOR(17 downto 0) := (others => '0');
		iTraceTrigCpu0 : in STD_LOGIC := '0';

		iTraceClkCpu1  : in STD_LOGIC := '0';
		iTraceDataCpu1 : in STD_LOGIC_VECTOR(17 downto 0) := (others => '0');
		iTraceTrigCpu1 : in STD_LOGIC := '0';

		iTraceClkCpu2  : in STD_LOGIC := '0';
		iTraceDataCpu2 : in STD_LOGIC_VECTOR(17 downto 0) := (others => '0');
		iTraceTrigCpu2 : in STD_LOGIC := '0';

		iTraceClkCpu3  : in STD_LOGIC := '0';
		iTraceDataCpu3 : in STD_LOGIC_VECTOR(17 downto 0) := (others => '0');
		iTraceTrigCpu3 : in STD_LOGIC := '0';

		oTraceClk  : out STD_LOGIC;
		oTraceData : out STD_LOGIC_VECTOR(17 downto 0);
		oTraceTrig : out STD_LOGIC

	);
end TraceSldNode;

architecture behaviour of TraceSldNode is

signal rIrReg    : STD_LOGIC_VECTOR(1 downto 0);
signal rShiftReg : STD_LOGIC_VECTOR(3 downto 0);

signal rSelReg   : STD_LOGIC_VECTOR(1 downto 0);
signal rSelSyncA : STD_LOGIC_VECTOR(1 downto 0);
signal rSelSyncB : STD_LOGIC_VECTOR(1 downto 0);

signal rTraceClk : STD_LOGIC;
signal rTraceTrig: STD_LOGIC;
signal rTraceData: STD_LOGIC_VECTOR(17 downto 0);

begin
	ir_out<=(others => '0');
	tdo<=rShiftReg(0);

	pJtagRegs:
	process(clrn,raw_tck)
	begin
		if (clrn='0') then
			rIrReg<=(others => '1');
			rShiftReg<=(others => '0');
			rSelReg<=(others => '0');
		elsif rising_edge(raw_tck) then
			if ena='1' then
				rIrReg<=ir_in;
			end if;

			-- Shift register is set in "capture DR" state
			-- Shift register is shifted in "Shift-DR" state
			if usr1='0' and ena='1' and (jtag_state_cdr='1' or jtag_state_sdr='1') then
				if jtag_state_cdr='1' then
					rShiftReg<=(others => '0');
					if rIrReg="10" then
						rShiftReg(rSelReg'range)<=rSelReg;
					end if;
				else
					rShiftReg(rShiftReg'high)<=tdi;
					rShiftReg(rShiftReg'high - 1 downto 0)<=rShiftReg(rShiftReg'high downto 1);
				end if;
			end if;

			-- Instruction "10" will write to the Trace selector
			if usr1='0' and ena='1' and jtag_state_udr='1' and rIrReg="10" then
				rSelReg<=rShiftReg(rSelReg'range);
			end if;
		end if;
	end process;

	oTraceClk<=rTraceClk;
	oTraceData<=rTraceData;
	oTraceTrig<=rTraceTrig;
	pMultiplex:
	process(iClkX2)
	begin
		if rising_edge(iClkX2) then
			rSelSyncA<=rSelReg;
			rSelSyncB<=rSelSyncA;
			case rSelSyncB is
				when "00" =>
					rTraceClk <= iTraceClkCpu0;
					rTraceData<= iTraceDataCpu0;
					rTraceTrig<= iTraceTrigCpu0;
				when "01" =>
					rTraceClk <= iTraceClkCpu1;
					rTraceData<= iTraceDataCpu1;
					rTraceTrig<= iTraceTrigCpu1;
				when "10" =>
					rTraceClk <= iTraceClkCpu2;
					rTraceData<= iTraceDataCpu2;
					rTraceTrig<= iTraceTrigCpu2;
				when others =>
					rTraceClk <= iTraceClkCpu3;
					rTraceData<= iTraceDataCpu3;
					rTraceTrig<= iTraceTrigCpu3;
			end case;
		end if;
	end process;
end behaviour;


-------------------------------------------------------------------
-- Toplevel module
-------------------------------------------------------------------
-- Include this module into your design.

-- Connect the iTraceClkCpuX,iTraceDataCpuX,iTraceTrigCpuX signals
-- to the various CPU trace outputs.

-- Connect oTraceClk,oTraceData,oTraceTrig to your external chip pins.
library IEEE;
use IEEE.std_logic_1164.all;

entity tracemultiplexer is
	port (
		iClkX2         : in STD_LOGIC;

		iTraceClkCpu0  : in STD_LOGIC := '0';
		iTraceDataCpu0 : in STD_LOGIC_VECTOR(17 downto 0) := (others => '0');
		iTraceTrigCpu0 : in STD_LOGIC := '0';

		iTraceClkCpu1  : in STD_LOGIC := '0';
		iTraceDataCpu1 : in STD_LOGIC_VECTOR(17 downto 0) := (others => '0');
		iTraceTrigCpu1 : in STD_LOGIC := '0';

		iTraceClkCpu2  : in STD_LOGIC := '0';
		iTraceDataCpu2 : in STD_LOGIC_VECTOR(17 downto 0) := (others => '0');
		iTraceTrigCpu2 : in STD_LOGIC := '0';

		iTraceClkCpu3  : in STD_LOGIC := '0';
		iTraceDataCpu3 : in STD_LOGIC_VECTOR(17 downto 0) := (others => '0');
		iTraceTrigCpu3 : in STD_LOGIC := '0';

		oTraceClk  : out STD_LOGIC;
		oTraceData : out STD_LOGIC_VECTOR(17 downto 0);
		oTraceTrig : out STD_LOGIC

	);

	attribute useioff : boolean;
	attribute useioff of oTraceClk  : signal is true;
	attribute useioff of oTraceData : signal is true;
	attribute useioff of oTraceTrig : signal is true;

end tracemultiplexer;

architecture behaviour of tracemultiplexer is

attribute preserve : boolean;

component TraceSldNode is
	generic (
		SLD_NODE_INFO : INTEGER := 16#08084C00#
	);
	port (

		iClkX2         : in STD_LOGIC;

		iTraceClkCpu0  : in STD_LOGIC := '0';
		iTraceDataCpu0 : in STD_LOGIC_VECTOR(17 downto 0) := (others => '0');
		iTraceTrigCpu0 : in STD_LOGIC := '0';

		iTraceClkCpu1  : in STD_LOGIC := '0';
		iTraceDataCpu1 : in STD_LOGIC_VECTOR(17 downto 0) := (others => '0');
		iTraceTrigCpu1 : in STD_LOGIC := '0';

		iTraceClkCpu2  : in STD_LOGIC := '0';
		iTraceDataCpu2 : in STD_LOGIC_VECTOR(17 downto 0) := (others => '0');
		iTraceTrigCpu2 : in STD_LOGIC := '0';

		iTraceClkCpu3  : in STD_LOGIC := '0';
		iTraceDataCpu3 : in STD_LOGIC_VECTOR(17 downto 0) := (others => '0');
		iTraceTrigCpu3 : in STD_LOGIC := '0';

		oTraceClk  : out STD_LOGIC;
		oTraceData : out STD_LOGIC_VECTOR(17 downto 0);
		oTraceTrig : out STD_LOGIC

	);
end component;

signal wTraceClk     : STD_LOGIC;
signal wTraceData    : STD_LOGIC_VECTOR(17 downto 0);
signal wTraceTrig    : STD_LOGIC;

-- ensure that rTraceClkTmp REALLY is instantiated by the synthesis.
signal rTraceClkTmp  : STD_LOGIC;
attribute preserve of rTraceClkTmp : signal is true;

signal rTraceClk     : STD_LOGIC;
signal rTraceData    : STD_LOGIC_VECTOR(17 downto 0);
signal rTraceTrig    : STD_LOGIC;

begin
	cMultiplex : TraceSldNode
	generic map (
		SLD_NODE_INFO => 16#08084C00#
	)
	port map (
		iClkX2 => iClkX2,

		iTraceClkCpu0  => iTraceClkCpu0,
		iTraceDataCpu0 => iTraceDataCpu0,
		iTraceTrigCpu0 => iTraceTrigCpu0,

		iTraceClkCpu1  => iTraceClkCpu1,
		iTraceDataCpu1 => iTraceDataCpu1,
		iTraceTrigCpu1 => iTraceTrigCpu1,

		iTraceClkCpu2  => iTraceClkCpu2,
		iTraceDataCpu2 => iTraceDataCpu2,
		iTraceTrigCpu2 => iTraceTrigCpu2,

		iTraceClkCpu3  => iTraceClkCpu3,
		iTraceDataCpu3 => iTraceDataCpu3,
		iTraceTrigCpu3 => iTraceTrigCpu3,

		oTraceClk   => wTraceClk,
		oTraceData  => wTraceData,
		oTraceTrig  => wTraceTrig
	);



	oTraceClk <=rTraceClk;
	oTraceData<=rTraceData;
	oTraceTrig<=rTraceTrig;

	pTraceClk:
	process(iClkX2)
	begin
		-- Use iClkX2 to delay trace clock by one and a half cycle.
		-- This way the trace clock is center aligned to the data eyes.
		if falling_edge(iClkX2) then
			-- The usage of rTraceClkTmp is INTENTIONAL:
			-- The rTraceClkTmp register is fed directly by the output
			-- register of the SOPC block. Now this SOPC output register
			-- is clocked on the rising edge of iClkX2.
			-- So you have only 1/2 clock cycle to transfer the content of this
			-- output register to rTraceClkTmp.
			-- Because we use rTraceClkTmp, it can be fitted right next to
			-- the SOPC output register and so timing requirements are met easily.
			-- rTraceClk resides in the IO element, but for the transfer from rTraceClkTmp
			-- to rTraceClk one full clock cycle is available.
			rTraceClkTmp <= wTraceClk;
			rTraceClk    <= rTraceClkTmp;
		end if;
	end process;

	pTraceData:
	process(iClkX2)
	begin
		if rising_edge(iClkX2) then
			rTraceData <= wTraceData;
			rTraceTrig <= wTraceTrig;
		end if;
	end process;
end behaviour;
